#!/usr/bin/env python

import argparse
import enum
import math
import os
import socket
from string import Template as PyTemplate
from typing import Optional, Sequence, TypedDict

parser = argparse.ArgumentParser()
parser.add_argument("--program", help="Program to run. e.g. ./lbm_d3q27", type=str)
parser.add_argument("--name", help="Run name", type=str)
parser.add_argument("--model", help="Path to model to use for experiment", type=str)
parser.add_argument(
    "--mem", help="Slurm memory limit for all nodes total. In GB", type=int
)
parser.add_argument("-n", help="Node count in format start:end", type=str)
parser.add_argument("--model-size", help="Model size.", type=int)
parser.add_argument("--generator", help="Generator to use", type=str)
parser.add_argument("-D", help="Dry-run", action="store_true")
parser.add_argument("--args", help="Args to pass to program", type=str)
parser.add_argument(
    "--override-data-dir",
    help="Use existing data-dir if exists",
    dest="override",
    action="store_true",
)
parser.add_argument(
    "--dloop",
    help="Enable loop optimizations",
    action="store_true",
)
parser.add_argument("--exports", help="Add extra exports. Comma separated", type=str)
parser.add_argument("--partition", help="Slurm partition", type=str)
parser.add_argument(
    "--max-tasks-per-node",
    help="Set maximum tasks per node",
    type=int,
    dest="max_tasks_per_node",
)

CLUSTER = "IDUN"

if "betzy" in socket.gethostname():
    CLUSTER = "betzy"

if "fram" in socket.gethostname():
    CLUSTER = "fram"

# Username for all clusers
USER = "username"

USERWORK = None
ACCOUNT = None
PARTITION = None
MIN_NODES = None
CORE_PER_NODE = None
DATA_DIR = None
BINVOX_DIR = None
CPU_PER_TASK = None
MAX_MEM_PER_NODE = 500

if CLUSTER == "IDUN":
    USERWORK = f"/cluster/work/{USER}"
    CORE_PER_NODE = 56
    MIN_NODES = 1
    ACCOUNT = ""
    PARTITION = "CPUQ"


if CLUSTER == "betzy" or CLUSTER == "fram":
    USERWORK = f"/cluster/work/users/{USER}"
    SIGMA_ACCOUNT = ""
    ACCOUNT = SIGMA_ACCOUNT
    CORE_PER_NODE = 128
    PARTITION = "normal"
    BINVOX_DIR = os.path.join(USERWORK, "binvox")
    MIN_NODES = 4
    MAX_MEM_PER_NODE = 248

if CLUSTER == "fram":
    MIN_NODES = 1
    CORE_PER_NODE = 32
    CPU_PER_TASK = 8
    MAX_MEM_PER_NODE = 58


DATA_DIR = f"/cluster/home/{USER}/lbm-d3-data"

TEMPLATE_DIR = f"{USERWORK}/lbm_d3"
TEMPLATE_FILE = "./lbm-d3.slurm"
WORKDIR = f"{USERWORK}/lbm-d3-jobs"


class Models(str, enum.Enum):
    RCA80 = "RCA_Solid_d80.binvox"
    RCA125 = "RCA_Solid_d125.binvox"
    RCA200 = "RCA_Solid_d200.binvox"
    RCA400 = "RCA_Solid_d400.binvox"

    P10400 = "p10-400.binvox"
    P20400 = "p20-400.binvox"
    P30400 = "p30-400.binvox"
    P40400 = "p40-400.binvox"
    P50400 = "p50-400.binvox"
    P60400 = "p60-400.binvox"
    P70400 = "p70-400.binvox"
    P80400 = "p80-400.binvox"
    P90400 = "p90-400.binvox"
    P100400 = "p100-400.binvox"

    NONE = ""


def mkdir_p(dir: str) -> None:
    if not os.path.exists(dir):
        os.mkdir(dir)


class Template(PyTemplate):
    """A template class using %-delimited vars"""

    delimiter = "%"


make_files = list(
    map(
        lambda f: os.path.join(TEMPLATE_DIR, f),
        ["lbm_*.c", "Makefile", "read_binvox.c", "read_binvox.h"],
    )
)


def create_job_dir(dirname: str) -> str:
    mkdir_p(WORKDIR)
    abs_dir = os.path.join(WORKDIR, dirname)
    mkdir_p(abs_dir)
    os.system(f"cp {' '.join(make_files)} {abs_dir}")
    return abs_dir


def create_data_dir(name: str, override: bool = False) -> str:
    orig_dir = os.path.join(DATA_DIR, name)
    dir = orig_dir

    if not override:
        i = 1
        while os.path.exists(dir):
            dir = f"{orig_dir}_{i}"
            i += 1

    mkdir_p(dir)
    return dir


class Job(TypedDict):
    partition: str
    nodes: int
    tasks_per_node: int
    cpus_per_task: int
    job_name: str
    out_file: str
    exports: str
    preface: str
    model_name: Optional[str]
    # Memory in
    mem: str
    account: Optional[str]


def get_slurm_template() -> Template:
    with open(os.path.join(TEMPLATE_DIR, TEMPLATE_FILE), "r") as f:
        src = Template(f.read())
    return src


def gen_perf_jobs(*, ranks: int = 4, data_dir: str) -> Sequence[Job]:
    cpus_per_task = 8 if CPU_PER_TASK is None else CPU_PER_TASK
    max_tasks_per_node = math.floor(CORE_PER_NODE / cpus_per_task)

    nodes = max(MIN_NODES, math.ceil(ranks / max_tasks_per_node))
    tasks_per_node = math.floor(ranks / nodes)

    models = [
        Models.RCA400,
        Models.P100400,
        Models.P80400,
        Models.P60400,
        Models.P10400,
    ]
    versions = [
        {
            "name": "base",
            "program": "lbm_d3q27_baseline_complex",
            "optimizations": [],
            "extra_args": "",
        },
        {
            "name": "IA",
            "program": "lbm_d3q27",
            "optimizations": [],
            "extra_args": "",
        },
        {
            "name": "IALO",
            "program": "lbm_d3q27",
            "optimizations": ["OSTREAM_LOOP_INDIRECT", "OCOLLIDE_LOOP_INDIRECT"],
            "extra_args": "",
        },
        {
            "name": "IALO1D",
            "program": "lbm_d3q27",
            "optimizations": ["OSTREAM_LOOP_INDIRECT", "OCOLLIDE_LOOP_INDIRECT"],
            "extra_args": " --cart-dimsx 1",
        },
    ]

    jobs: list[Job] = []

    for version in versions:
        for model in models:
            model_name = model.replace(".binvox", "")
            perf_file = os.path.join(
                data_dir, f"lbm-perf-{version['name']}-{model_name}.csv"
            )
            jobs.append(
                {
                    "partition": "normal",
                    "account": ACCOUNT,
                    "nodes": nodes,
                    "tasks_per_node": tasks_per_node,
                    "cpus_per_task": cpus_per_task,
                    "job_name": f"lbm-d3-{model_name}-{version['name']}",
                    "model_name": f"{model}",
                    "out_file": "job.out",
                    "exports": f"export OFLAGS=\"-D {' -D '.join(version['optimizations'])}\"\n",
                    "preface": f'PROG={version["program"]}'
                    + f"\nUSERWORK={USERWORK}"
                    + f'\nEXTRA_ARGS=" --perf-file {perf_file}"\n'
                    + f"\nEXTRA_ARGS+= {version['extra_args']}",
                    "mem": "40G",
                }
            )

    return jobs


def generate_jobs(
    *,
    model: Models,
    optimizations: list[str],
    range_start: int = 4,
    range_end: int = 32,
    range_step: int = 4,
    data_dir: str,
    max_tasks_per_node_o: int,
) -> Sequence[Job]:
    cpus_per_task = 8 if CPU_PER_TASK is None else CPU_PER_TASK
    max_tasks_per_node = math.floor(CORE_PER_NODE / cpus_per_task)
    if max_tasks_per_node_o is not None:
        max_tasks_per_node = max_tasks_per_node_o

    print(
        f"Cluster configuration:\tTask/Node: {max_tasks_per_node}\tCPU/Task: {cpus_per_task}"
    )

    jobs: list[Job] = []

    for i in range(range_start, range_end + 1, range_step):
        nodes = max(MIN_NODES, math.ceil(i / max_tasks_per_node))
        tasks_per_node = math.floor(i / nodes)
        total_tasks = tasks_per_node * nodes
        perf_file = os.path.join(data_dir, f"lbm-perf-{total_tasks}.csv")

        jobs.append(
            {
                "partition": PARTITION,
                "account": ACCOUNT,
                "nodes": nodes,
                "tasks_per_node": tasks_per_node,
                "cpus_per_task": cpus_per_task,
                "job_name": f"lbm-d3-r{total_tasks}-n{nodes}-t{tasks_per_node}",
                "out_file": "job.out",
                "model_name": model,
                "exports": f"export OFLAGS=\"-D {' -D '.join(optimizations)}\"\n",
                "preface": f'USERWORK={USERWORK}\nEXTRA_ARGS=" --perf-file {perf_file}"\n',
                "mem": "40G",
            }
        )

    return jobs


def gen_usage_jobs(
    *,
    optimizations: list[str],
    range_start: int = 1,
    range_end: int = 8,
    step: int = 1,
    model_size: int = 1024,
    data_dir: str,
    tasks: int = 16,
) -> Sequence[Job]:
    cpus_per_task = 8 if CPU_PER_TASK is None else CPU_PER_TASK
    max_tasks_per_node = math.floor(CORE_PER_NODE / cpus_per_task)

    nodes = max(MIN_NODES, math.ceil(tasks / max_tasks_per_node))
    tasks_per_node = math.floor(tasks / nodes)

    jobs: list[Job] = []
    for i in range(range_start, range_end, step):
        perf_file = os.path.join(data_dir, f"lbm-perf-{i}.csv")

        jobs.append(
            {
                "partition": PARTITION,
                "account": ACCOUNT,
                "nodes": 4,
                "tasks_per_node": tasks_per_node,
                "cpus_per_task": cpus_per_task,
                "job_name": f"lbm-d3-i{i}-t{tasks_per_node}-c{cpus_per_task}",
                "model_name": f"p{i}-{model_size}.binvox",
                "out_file": "job.out",
                "exports": f"export OFLAGS=\"-D {' -D '.join(optimizations)}\"\n",
                "preface": f'USERWORK={USERWORK}\nEXTRA_ARGS=" --perf-file {perf_file}"\n',
                "mem": "40G",
            }
        )

    return jobs


def submit_jobs(args):
    start = 4
    end = 32
    if args.n:
        start = int(args.n.split(":")[0])
        end = int(args.n.split(":")[1])

    template = get_slurm_template()

    data_dir = create_data_dir(args.name, args.override)
    print("Created data dir: ", data_dir)

    optimizations = []

    if args.dloop:
        optimizations = ["OSTREAM_LOOP_INDIRECT", "OCOLLIDE_LOOP_INDIRECT"]
        print("Loop optimizations ON")

    if args.model:
        model_path = os.path.join(BINVOX_DIR, args.model)
        if not os.path.exists(model_path):
            print(f"Could not read model file: {model_path}")
            print("Ensure that the model exists and that you entered a valid name")
            exit(1)

    jobs = []
    if args.generator == "usage":
        jobs = gen_usage_jobs(
            optimizations=optimizations,
            data_dir=data_dir,
            range_start=10,
            range_end=110,
            step=10,
            model_size=args.model_size if args.model_size > 0 else 1024,
        )

    elif args.generator == "scale":
        model = Models.RCA80
        if args.model:
            model = args.model

        jobs = generate_jobs(
            model=model,
            data_dir=data_dir,
            # It seems the loop optimizations actually make it worse...
            # Still an interesting result
            # Contra! Much faster!! ???
            optimizations=optimizations,
            range_start=start,
            range_end=end,
            max_tasks_per_node_o=args.max_tasks_per_node,
        )

    elif args.generator == "general":
        jobs = gen_perf_jobs(ranks=start, data_dir=data_dir)

    print("--------------")

    for job in jobs:
        if args.mem:
            node_mem = math.ceil(args.mem / job["nodes"])
            job["mem"] = f"{node_mem}G"

        model_name = job["model_name"]
        dir_prefix = (
            (
                args.program.replace("./", "")
                if args.program is not None
                else "lbm_d3q27"
            )
            + "_"
            + model_name.replace(".binvox", "")
            + "_"
        )
        if args.name:
            dir_prefix = args.name
        job_dir = create_job_dir(dir_prefix + job["job_name"])
        job["out_file"] = os.path.join(job_dir, "job.out")
        job["preface"] += f"WORKDIR={job_dir}\n"

        if args.program is not None:
            job["preface"] += f"PROG={args.program}\n"

        if args.args is not None:
            job["preface"] += f'EXTRA_ARGS+=" {args.args}"\n'

        if args.exports is not None:
            for item in args.exports.split(","):
                job["exports"] += f"export {item}\n"

        if args.partition is not None:
            job["partition"] = args.partition

            if args.partition == "bigmem":
                ranks = job["tasks_per_node"] * job["nodes"]
                mem_per_rank = args.mem / ranks
                mem_per_node = args.mem / job["nodes"]
                if mem_per_node > MAX_MEM_PER_NODE:
                    mem_per_cpu = math.ceil(mem_per_rank / job["cpus_per_task"])
                    job["preface"] = (
                        f"#SBATCH --mem-per-cpu={mem_per_cpu}G\n" + job["preface"]
                    )
                    job["nodes"] = math.floor(job["nodes"] / 2)
                    job["tasks_per_node"] = math.ceil(job["tasks_per_node"] * 2)
                else:
                    job["partition"] = PARTITION

        slurm_file_str = template.substitute(job)
        slurm_file = os.path.join(job_dir, "job.slurm")

        with open(slurm_file, "w") as f:
            f.write(slurm_file_str)
        print(
            f"Submitting job {job['job_name']}"
            + f"\tN: {job['nodes']}"
            + f"\tranks/node: {job['tasks_per_node']}"
            + f"\tMEM: {job['mem']}"
            + f"\tRanks: {job['tasks_per_node'] * job['nodes']}"
            + f"\tTotal CPU count: {job['tasks_per_node'] * job['nodes'] * job['cpus_per_task']}"
        )
        if not args.D:
            os.system(f"sbatch {slurm_file}")
        else:
            print("[DRY RUN] skipping sbatch...")


def main():
    args = parser.parse_args()
    submit_jobs(args)


if __name__ == "__main__":
    main()
